package nondeterministic_fixture_test

import (
	. "github.com/onsi/ginkgo/v2"
)

var cheat = func() {
	It("in", func() {

	})

	It("order", func() {

	})
}

var _ = Describe("ordered", Ordered, func() {
	It("always", func() {

	})

	It("runs", func() {

	})

	cheat()
})
