package malformed_fixture_test

import (
	"testing"

	. "github.com/onsi/ginkgo/v2"
	. "github.com/onsi/gomega"
)

func TestMalformedFixture(t *testing.T) {
	RegisterFailHandler(Fail)
	RunSpecs(t, "MalformedFixture Suite")
}

var _ = Describe("Malformed by...", func() {
	By("right here...")

	It("should never run", func() {

	})
})
